#include "scenes/combat_scene.h"

#include "worlds/combat_world.h"
#include "gfx.h"
#include "engine.h"
#include "save.h"

#include <pystring.h>

using namespace engine;

namespace mancers
{
    struct TerrainVertex
    {
        glm::vec4 m_position = glm::vec4(0.0f);
        glm::vec3 m_texcoord = glm::vec3(0.0f);
        glm::vec3 m_normal = glm::vec3(0.0f);
    };

    namespace e_terrainGenFunc
    {
        enum terrainGenFunc
        {
            standard,
            ridge,
            valley
        };
    }
    typedef e_terrainGenFunc::terrainGenFunc GenFunc;

    struct GridInfo
    {
        float*  m_heights = nullptr;
        u32*    m_surfaceIndices = nullptr;
    };

    struct MapGenParams
    {
        glm::vec3                   m_gridToUnit = glm::vec3(1.0f);
        glm::vec3                   m_totalGridSize = glm::vec3(1.0f);
        glm::vec3                   m_playableGridSize = glm::vec3(1.0f);
        float                       m_seed = 0.0f;
        float                       m_frequency = 1.0f;
        u32                         m_numGroups = 0;
        std::vector<std::string>    m_textureFilepaths;
    };

    struct MapGenOutput
    {
        std::vector<glm::vec3>  m_groupPositions;
        GridInfo                m_grid = {};
    };

    struct MapGenData
    {
        MapGenParams    m_params = {};
        MapGenOutput*   m_output = nullptr;
    };

    struct GroupWeightedSpawn
    {
        std::string m_name = "";
        float       m_lower = 0.0f;
        float       m_upper = 1.0f;
    };

	Scene* createCombatScene(void* data)
	{
		Scene* scene = memNew<CombatScene>();
		return scene;
	}

    glm::vec3 calculateTriangleNormal(const glm::vec3& a, const glm::vec3& b, const glm::vec3& c)
    {
        glm::vec3 ab = b - a;
        glm::vec3 ac = c - a;

        return glm::cross(ab, ac) * -1.0f;
    }


    ModelAsyncData generateMap(void* data)
    {        
        MapGenData* d = (MapGenData*)data;
        const MapGenParams& params = d->m_params;
        MapGenOutput* genOut = d->m_output;
        const glm::vec3& genGridSize = params.m_totalGridSize;
        const glm::vec3& gridToUnit = params.m_gridToUnit;
        
        float* heights= (float*)memAlloc(sizeof(float) * (size_t)genGridSize.z * (size_t)genGridSize.x);
        memset(heights, 0, sizeof(float) * (size_t)genGridSize.z * (size_t)genGridSize.x);

        u32* surfaces = (u32*)memAlloc(sizeof(u32) * (size_t)genGridSize.z * (size_t)genGridSize.x);
        memset(surfaces, 0, sizeof(u32) * (size_t)genGridSize.z * (size_t)genGridSize.x);

        // generate some heights
        for (size_t z = 0; z < (size_t)genGridSize.z; ++z)
        {
            for (size_t x = 0; x < (size_t)genGridSize.x; ++x)
            {
                size_t i = z * (size_t)genGridSize.x + x;

                glm::vec2 st =
                {
                    ((float)x / genGridSize.x) * params.m_frequency,
                    ((float)z / genGridSize.z) * params.m_frequency
                };

                float h = maths::generateFbmNoise(st, params.m_seed);
                heights[i] = floor(h * genGridSize.y * 2.0f) * 0.5f;

                surfaces[i] = (u32)std::round(maths::generateFbmNoise(st, params.m_seed * 2.0f));
            }
        }

        // distribute groups around the playable grid area
        for (u32 i = 0; i < params.m_numGroups; ++i)
        {
            glm::vec3 pos = glm::vec3(0.0f);

            bool valid = false;
            while (!valid)
            {
                pos.x = 1.0f + (rand() % (int)(params.m_playableGridSize.x - 2.0f));
                pos.z = 1.0f + (rand() % (int)(params.m_playableGridSize.z - 2.0f));
                
                valid = true;

                // check for overlay with other groups
                for (auto& p : genOut->m_groupPositions)
                {
                    if (abs(pos.x - p.x) < 3.0f || abs(pos.z - p.z) < 3.0f)
                    {
                        valid = false;
                        break;
                    }
                }
            }
            genOut->m_groupPositions.push_back(pos);
        }

        // flatten area around group positions
        glm::vec3 boundary = (params.m_totalGridSize - params.m_playableGridSize) * 0.5f;
        for (auto& p : genOut->m_groupPositions)
        {
            size_t width = (size_t)genGridSize.x;
            size_t x = (size_t)(p.x + boundary.x);
            size_t z = (size_t)(p.z + boundary.z);
            size_t i = x + (z * width);

            float height = heights[i];

            heights[(i - 1) - width] = height;
            heights[i - width] = height;
            heights[(i + 1) - width] = height;
            heights[i - 1] = height;
            heights[i + 1] = height;
            heights[(i - 1) + width] = height;
            heights[i + width] = height;
            heights[(i + 1) + width] = height;

            p.y = height;
        }

        std::vector<TerrainVertex> verts;
        std::vector<u32> inds;

        glm::vec3 pos = glm::vec3(-genGridSize.x * 0.5f, 0.0f, -genGridSize.z * 0.5f);

        int indexOffsets[4] =
        {
             1,
            -1,
            (int)genGridSize.x,
            (int)-genGridSize.x
        };
            
        glm::vec3 offsetNormals[4] =
        {
            glm::vec3(1.0f, 0.0f, 0.0f),
            glm::vec3(-1.0f, 0.0f, 0.0f),
            glm::vec3(0.0f, 0.0f, 1.0f),
            glm::vec3(0.0f, 0.0f, -1.0f)
        };

        glm::uvec2 offsetCorners[4] =
        {
            glm::uvec2(1, 2),
            glm::uvec2(0, 3),
            glm::uvec2(2, 0),
            glm::uvec2(3, 1)
        };

        for (float z = 0.0f; z < genGridSize.z; z += 1.0f)
        {
            for (float x = 0.0f; x < genGridSize.x; x += 1.0f)
            {
                size_t i = (size_t)z * (size_t)genGridSize.x + (size_t)x;
                float height = heights[i];

                glm::vec2 st =
                {
                    ((float)x / genGridSize.x) * params.m_frequency,
                    ((float)z / genGridSize.z) * params.m_frequency
                };

                float materialIndex = (float)surfaces[i];
                
                glm::vec3 corners[4];

                {
                    size_t i0 = i;
                    size_t i1 = i;
                    size_t i2 = i;
                    size_t i3 = i;

                    corners[0].x = pos.x * gridToUnit.x;
                    corners[0].y = heights[i0] * gridToUnit.y;
                    corners[0].z = (pos.z + 1.0f) * gridToUnit.z;

                    // create the main grid face
                    TerrainVertex v0 = {};
                    v0.m_position = glm::vec4(corners[0], 1.0f);
                    v0.m_texcoord = glm::vec3(0.0f, 0.0f, materialIndex);

                    corners[1].x = (pos.x + 1.0f) * gridToUnit.x;
                    corners[1].y = heights[i1] * gridToUnit.y;
                    corners[1].z = pos.z * gridToUnit.z;

                    TerrainVertex v1 = {};
                    v1.m_position = glm::vec4(corners[1], 1.0f);
                    v1.m_texcoord = glm::vec3(1.0f, 1.0f, materialIndex);

                    corners[2].x = (pos.x + 1.0f) * gridToUnit.x;
                    corners[2].y = heights[i2] * gridToUnit.y;
                    corners[2].z = (pos.z + 1.0f) * gridToUnit.z;
                   
                    TerrainVertex v2 = {};
                    v2.m_position = glm::vec4(corners[2], 1.0f);
                    v2.m_texcoord = glm::vec3(1.0f, 0.0f, materialIndex);

                    corners[3].x = pos.x * gridToUnit.x;
                    corners[3].y = heights[i3] * gridToUnit.y;
                    corners[3].z = pos.z * gridToUnit.z;

                    TerrainVertex v3 = {};
                    v3.m_position = glm::vec4(corners[3], 1.0f);
                    v3.m_texcoord = glm::vec3(0.0f, 1.0f, materialIndex);

                    // determine surface normals of triangles
                    glm::vec3 normalA = calculateTriangleNormal(v0.m_position, v1.m_position, v2.m_position);
                    glm::vec3 normalB = calculateTriangleNormal(v0.m_position, v3.m_position, v1.m_position);

                    v0.m_normal = (normalA + normalB) * 0.5f;
                    v1.m_normal = (normalA + normalB) * 0.5f;
                    v2.m_normal = normalA;
                    v3.m_normal = normalB;

                    u32 baseVert = (u32)verts.size();
                    verts.push_back(v0);
                    verts.push_back(v1);
                    verts.push_back(v2);
                    verts.push_back(v3);

                    inds.push_back(baseVert);
                    inds.push_back(baseVert + 1);
                    inds.push_back(baseVert + 2);
                    inds.push_back(baseVert);
                    inds.push_back(baseVert + 3);
                    inds.push_back(baseVert + 1);
                }

                // check if any adjacent heights are lower and fill in squares if they are
                for (int j = 0; j < 4; ++j)
                {
                    int a = (int)i + indexOffsets[j];
                    if (a < 0 || a >= (int)(genGridSize.x * genGridSize.z))
                        continue;

                    float adjacent = heights[a];
                    if (adjacent >= height)
                        continue;

                    TerrainVertex v0 = {};
                    v0.m_position.x = corners[offsetCorners[j].x].x;
                    v0.m_position.y = height;
                    v0.m_position.z = corners[offsetCorners[j].x].z;
                    v0.m_position.w = 1.0f;
                    v0.m_texcoord = glm::vec3(0.0f, 0.0f, materialIndex);
                    v0.m_normal = offsetNormals[j];

                    TerrainVertex v1 = {};
                    v1.m_position.x = corners[offsetCorners[j].y].x;
                    v1.m_position.y = adjacent;
                    v1.m_position.z = corners[offsetCorners[j].y].z;
                    v1.m_position.w = 1.0f;
                    v1.m_texcoord = glm::vec3(1.0f, 1.0f, materialIndex);
                    v1.m_normal = offsetNormals[j];

                    TerrainVertex v2 = {};
                    v2.m_position.x = corners[offsetCorners[j].y].x;
                    v2.m_position.y = height;
                    v2.m_position.z = corners[offsetCorners[j].y].z;
                    v2.m_position.w = 1.0f;
                    v2.m_texcoord = glm::vec3(1.0f, 0.0f, materialIndex);
                    v2.m_normal = offsetNormals[j];

                    TerrainVertex v3 = {};
                    v3.m_position.x = corners[offsetCorners[j].x].x;
                    v3.m_position.y = adjacent;
                    v3.m_position.z = corners[offsetCorners[j].x].z;
                    v3.m_position.w = 1.0f;
                    v3.m_texcoord = glm::vec3(0.0f, 1.0f, materialIndex);
                    v3.m_normal = offsetNormals[j];

                    u32 baseVert = (u32)verts.size();
                    verts.push_back(v0);
                    verts.push_back(v1);
                    verts.push_back(v2);
                    verts.push_back(v3);

                    inds.push_back(baseVert);
                    inds.push_back(baseVert + 1);
                    inds.push_back(baseVert + 2);
                    inds.push_back(baseVert);
                    inds.push_back(baseVert + 3);
                    inds.push_back(baseVert + 1);

                }

                pos.x += 1.0f;
            }

            pos.x = -genGridSize.x * 0.5f;
            pos.z += 1.0f;
        }

        GeometryCpuData geom = {};
        geom.m_numVertices = (u32)verts.size();
        geom.m_vertexData.m_size = sizeof(TerrainVertex) * verts.size();
        geom.m_vertexData.m_data = memAlloc(geom.m_vertexData.m_size);
        memcpy(geom.m_vertexData.m_data, verts.data(), geom.m_vertexData.m_size);

        geom.m_numIndices = (u32)inds.size();
        geom.m_indexData.m_size = sizeof(u32) * inds.size();
        geom.m_indexData.m_data = memAlloc(geom.m_indexData.m_size);
        memcpy(geom.m_indexData.m_data, inds.data(), geom.m_indexData.m_size);

        geom.m_minVert[0] = FLT_MAX;
        geom.m_minVert[1] = FLT_MAX;
        geom.m_minVert[2] = FLT_MAX;
        geom.m_maxVert[0] = -FLT_MAX;
        geom.m_maxVert[1] = -FLT_MAX;
        geom.m_maxVert[2] = -FLT_MAX;

        for (auto& v : verts)
        {
            for (size_t i = 0; i < 3; ++i)
            {
                geom.m_minVert[i] = std::min(geom.m_minVert[i], v.m_position[i]);
                geom.m_maxVert[i] = std::max(geom.m_maxVert[i], v.m_position[i]);
            }
        }

        genOut->m_grid.m_heights = heights;
        genOut->m_grid.m_surfaceIndices = surfaces;

        ResourceManager* rm = ResourceManager::get();
        
        ModelAsyncData output = {};
        output.m_cpuBuffers.push_back(geom);

        MaterialLoadDesc matDesc = {};
        for (auto& s : params.m_textureFilepaths)
        {
            matDesc.m_textureFilenames.push_back(s);
        }
        matDesc.m_defaultSamplerID = hash<std::string>("clamp_linear");
        output.m_materials.push_back(rm->load<MaterialData, MaterialLoadDesc>(matDesc, e_resourceLoadFlags::async));

        return output;
    }

	void CombatScene::load(const char* fp)
	{
        if (m_flags & e_sceneFlags::loading)
            return;

        m_flags |= e_sceneFlags::loading;

        if (!utils::fileExists(fp))
        {
            m_flags &= ~e_sceneFlags::loading;
            return;
        }

        void* buf = nullptr;
        size_t len = 0;
        utils::readFileToBuffer(fp, buf, len);
        rapidjson::Document d;
        bool result = d.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)buf).HasParseError();
        if (result)
        {
            auto error = d.GetParseError();
            printf("Error!");
            m_flags &= ~e_sceneFlags::loading;
            ASSERT(false);
            return;
        }
        utils::freeFileBuffer(buf);

        // load prefabs into memory
        rapidjson::Value prefabs;
        prefabs.SetObject();
        if (d.HasMember("prefabs"))
        {
            prefabs = loadPrefabs(d["prefabs"], d.GetAllocator());
        }

        // load layers from prefab
        for (auto pit = prefabs.MemberBegin(); pit != prefabs.MemberEnd(); ++pit)
        {
            const rapidjson::Value& prefab = pit->value;
            if (prefab.HasMember("layer_order"))
            {
                const rapidjson::Value& loj = prefab["layer_order"].GetArray();
                for (size_t i = 0; i < loj.Size(); ++i)
                {
                    m_layerOrder.push_back(hash<std::string>(loj[i].GetString()));
                }
            }

            if (prefab.HasMember("layers"))
            {
                const rapidjson::Value& lj = prefab["layers"];
                for (auto& mit = lj.MemberBegin(); mit != lj.MemberEnd(); ++mit)
                {
                    const rapidjson::Value& layerJson = mit->value;
                    std::string layerName = mit->name.GetString();
                    ID layerID = hash(layerName);

                    m_layers[layerID] = loadLayer(layerJson);
                }
                break;
            }
        }

        m_world = memNew<CombatWorld>();
        CombatWorld& w = *(CombatWorld*)m_world;
        w.grow(32);

        ResourceManager* rm = ResourceManager::get();

        m_animManager = memNew<AnimationManager>();
        m_animManager->setup(m_world);

        // create ui cursor
        size_t uce = w.newEntity();
        loadEntity(uce, prefabs["common"]["entities"]["cursor"]);
        w.m_layer[uce] = hash<std::string>("cursor");
        w.m_scale[uce] = glm::vec3(0.01f, 0.02f, 1.0f);
        
        // create world cursor

        // generate map geometry
        size_t mge = w.newEntity();
        w.m_components[mge] |= (e_cmp::transform | e_cmp::modelRef | e_cmp::renderable);
        w.m_entityFlags[mge] |= e_entityFlags::passthroughCursor;
        w.m_layer[mge] = hash<std::string>("default");
        w.m_scale[mge] = glm::vec3(1.0f, 1.0f, 1.0f);
        w.m_pipeline[mge] = hash<std::string>("terrain");
        w.m_uniformData[mge].m_userData0 = glm::vec4(1.0f, 1.0f, 1.0f, 1.0f);

        MapGenParams mapGenParams = {};

        // load terrain gen parameters
        srand(utils::getSystemTimeMs());

        glm::vec3 boundarySize = glm::vec3(8.0f, 0.0f, 8.0f);
        glm::vec3 minGridSize = glm::vec3(1.0);
        glm::vec3 maxGridSize = glm::vec3(1.0);
        glm::vec3 gridToUnit = glm::vec3(1.0);

        getJsonArrayOrDefault(d["map_gen"]["grid_size"], "min", &minGridSize.x, 3, 1.0f);
        getJsonArrayOrDefault(d["map_gen"]["grid_size"], "max", &maxGridSize.x, 3, 1.0f);
        getJsonArrayOrDefault(d["map_gen"], "grid_to_unit_scale", &gridToUnit.x, 3, 1.0f);

        glm::vec3 genGridSize = glm::vec3(1.0);
        for (int i = 0; i < 3; ++i)
        {
            float r = ((float)rand() / RAND_MAX);
            genGridSize[i] = minGridSize[i] + floor((maxGridSize[i] - minGridSize[i]) * r);
            
            // enforce odd sizing on x and z for the purposes of generation
            if (i != 1 && (int)genGridSize[i] % 2 == 0)
                genGridSize[i] += 1.0f;

            mapGenParams.m_playableGridSize[i] = genGridSize[i];

            genGridSize[i] += boundarySize[i] * 2.0f;
        }
        
        mapGenParams.m_totalGridSize = genGridSize;
        mapGenParams.m_gridToUnit = gridToUnit;    

        w.m_gridScale[mge] = mapGenParams.m_playableGridSize;

        float minFrequency = getJsonValueOrDefault(d["map_gen"], "min_frequency", 1.0f);
        float maxFrequency = getJsonValueOrDefault(d["map_gen"], "max_frequency", 1.0f);
        mapGenParams.m_frequency = minFrequency + ((float)rand() / RAND_MAX) * (maxFrequency - minFrequency);

        // select number of NPC groups to generate
        u32 minNpcGroups = getJsonValueOrDefault<u32>(d["npc_gen"], "min_groups", 1);
        u32 maxNpcGroups = getJsonValueOrDefault<u32>(d["npc_gen"], "max_groups", 1);
        u32 numNpcGroups = minNpcGroups + (rand() % (maxNpcGroups - minNpcGroups));
        mapGenParams.m_numGroups = numNpcGroups + 1; // add an extra group for the player characters

        const rapidjson::Value& surfaces = d["map_gen"]["surfaces"];
        for (size_t i = 0; i < surfaces.Size(); ++i)
        {
            mapGenParams.m_textureFilepaths.push_back(getJsonValueOrDefault<const char*>(surfaces[i], "surface_texture", ""));
            mapGenParams.m_textureFilepaths.push_back(getJsonValueOrDefault<const char*>(surfaces[i], "edge_texture", ""));
        }

        MapGenOutput mapGenOutput = {};
        MapGenData* mapGenData = memNew<MapGenData>();
        mapGenData->m_params = mapGenParams;
        mapGenData->m_output = &mapGenOutput;

        ModelLoadDesc md = {};
        md.m_genFunc = generateMap;
        md.m_genData.m_size = sizeof(MapGenData);
        md.m_genData.m_data = mapGenData;
        w.m_modelRef[mge] = rm->load<ModelData, ModelLoadDesc>(md, e_resourceLoadFlags::async);
        
        // create border cube
        size_t be = w.newEntity();
        w.m_layer[be] = hash<std::string>("transparency");
        loadEntity(be, prefabs["combat"]["entities"]["border"]);
        w.m_gridScale[be] = mapGenParams.m_playableGridSize + glm::vec3(1.0f, 1.0f, 1.0f);
        w.m_gridScale[be].y *= 2.0f;
        w.m_scale[be] = w.m_gridScale[be] * mapGenParams.m_gridToUnit * 0.9999f;
        w.m_pos[be] = glm::vec3(0.5f, 0.0f, 0.5f);

        // create directional light
        size_t le = w.newEntity();
        w.m_layer[le] = hash<std::string>("default");
        loadEntity(le, prefabs["combat"]["entities"]["scene_directional_light"]);

        // create necessary cameras
        size_t pe = w.newEntity();
        loadEntity(pe, prefabs["combat"]["entities"]["perspective_camera_pivot"]);

        size_t oce = w.newEntity();
        loadEntity(oce, prefabs["combat"]["entities"]["ui_ortho_camera"]);

        size_t cce = w.newEntity();
        loadEntity(cce, prefabs["combat"]["entities"]["cursor_ortho_camera"]);

        // create skybox entity
        size_t sbe = w.newEntity();
        loadEntity(sbe, prefabs["combat"]["entities"]["skybox"]);
        w.m_layer[sbe] = hash<std::string>("skybox");

        // assign group placement points
        std::vector<int> groupAssignments;
        float furthest = -FLT_MAX;
        for (size_t i = 0; i < mapGenOutput.m_groupPositions.size(); ++i)
        {
            float closest = FLT_MAX;
            glm::vec2 a = glm::vec2(mapGenOutput.m_groupPositions[i].x, mapGenOutput.m_groupPositions[i].z);
            for (size_t j = 0; j < mapGenOutput.m_groupPositions.size(); ++j)
            {
                if (i == j)
                    continue;
                
                glm::vec2 b = glm::vec2(mapGenOutput.m_groupPositions[j].x, mapGenOutput.m_groupPositions[j].z);
                closest = std::min(closest, glm::length(b - a));
            }

            if (closest > furthest)
            {
                // assign furthest group away from others as player spawn
                for (auto& g : groupAssignments)
                    g = 0;
                groupAssignments.push_back(1);
                furthest = closest;
            }
            else
            {
                groupAssignments.push_back(0);
            }
        }

        // generate npc groups

        // map spawn rates for each group
        const rapidjson::Value& npcJson = d["npc_gen"];
        float totalSpawns = 0.0f;
        std::vector<GroupWeightedSpawn> weightedSpawns;
        for (auto git = npcJson["groups"].MemberBegin(); git != npcJson["groups"].MemberEnd(); ++git)
        {
            float spawnWeight = getJsonValueOrDefault(git->value, "spawn_chance", 1.0f);
            totalSpawns += spawnWeight;

            GroupWeightedSpawn spawn = {};
            spawn.m_name = git->name.GetString();
            spawn.m_upper = spawnWeight;
            weightedSpawns.push_back(spawn);
        }

        float current = 0.0f;
        for (auto& s : weightedSpawns)
        {
            float weighted = s.m_upper / totalSpawns;
            s.m_lower = current;
            s.m_upper = current + weighted;
            current += weighted;
        }

        glm::vec3 gridOffsets[9] = {
            { -1.0f, 0.0f, -1.0f },
            {  0.0f, 0.0f, -1.0f },
            {  1.0f, 0.0f, -1.0f },
            { -1.0f, 0.0f,  0.0f },
            {  0.0f, 0.0f,  0.0f },
            {  1.0f, 0.0f,  0.0f },
            { -1.0f, 0.0f,  1.0f },
            {  0.0f, 0.0f,  1.0f },
            {  1.0f, 0.0f,  1.0f },
        };

        for (size_t i = 0; i < mapGenOutput.m_groupPositions.size(); ++i)
        {
            if (groupAssignments[i] == 1)
                continue;

            glm::vec3 groupGridCentre = mapGenOutput.m_groupPositions[i];

            glm::vec3 groupCentre = glm::vec3(0.0f);
            groupCentre.x = ((genGridSize.x * -0.5f) + 0.5f + groupGridCentre.x + boundarySize.x) * gridToUnit.x;
            groupCentre.y = groupGridCentre.y * gridToUnit.y;
            groupCentre.z = ((genGridSize.z * -0.5f) + 0.5f + groupGridCentre.z + boundarySize.z) * gridToUnit.z;

            float spawnVal = (float)rand() / RAND_MAX;
            for (auto& s : weightedSpawns)
            {
                if (spawnVal > s.m_upper || spawnVal < s.m_lower)
                    continue;

                // spawn this group
                const rapidjson::Value& groupJson = npcJson["groups"][s.m_name.c_str()];

                int minSize = getJsonValueOrDefault(groupJson, "min_size", 1);
                int maxSize = getJsonValueOrDefault(groupJson, "max_size", 1);
                int size = (maxSize != minSize) ? minSize + (rand() % (maxSize - minSize)) : minSize;

                std::set<int> used;
                int numContains = groupJson["contains"].Size();
                for (int j = 0; j < size; ++j)
                {
                    int index = rand() % numContains;

                    std::string charPrefabName = groupJson["contains"][index].GetString();

                    std::vector<std::string> charPrefabAndName;
                    pystring::split(charPrefabName, charPrefabAndName, "/");

                    size_t e = w.newEntity();
                    loadEntity(e, prefabs[charPrefabAndName[0].c_str()]["entities"][charPrefabAndName[1].c_str()]);
                    w.m_entityFlags[e] |= e_combatFlags::interactive;
                    w.m_components[e] |= e_cmp::transform;
                    w.m_layer[e] = hash<std::string>("default");
                    w.m_scale[e] = gridToUnit * 0.8f;
                    
                    int g = rand() % 9;
                    while (!used.insert(g).second)
                    {
                        g = rand() % 9;
                    }

                    w.m_gridPos[e] = groupGridCentre + gridOffsets[g];

                    w.m_pos[e] = groupCentre + gridOffsets[g] * gridToUnit;
                    w.m_pos[e].y += gridToUnit.y * 0.4f;
                }
            }
        }

        // create entities for player characters
        Engine* e = Engine::get();
        GameState& state = e->getGameState();

        if (state["saved"].contains("party"))
        {
            size_t group = 0;
            for (group = 0; group < groupAssignments.size(); ++group)
            {
                if (groupAssignments[group] == 1)
                    break;
            }

            glm::vec3 groupGridCentre = mapGenOutput.m_groupPositions[group];

            glm::vec3 groupCentre = glm::vec3(0.0f);
            groupCentre.x = ((genGridSize.x * -0.5f) + 0.5f + groupGridCentre.x + boundarySize.x) * gridToUnit.x;
            groupCentre.y = groupGridCentre.y * gridToUnit.y;
            groupCentre.z = ((genGridSize.z * -0.5f) + 0.5f + groupGridCentre.z + boundarySize.z) * gridToUnit.z;

            // create the party from game state
            GameState& party = state["saved"]["party"];
            
            std::set<int> used;
            for (auto it = party.begin(); it != party.end(); ++it)
            {
                size_t e = w.newEntity();
                loadEntity(e, prefabs["combat"]["entities"]["player"]);
                w.m_entityFlags[e] |= e_combatFlags::interactive;

                int g = rand() % 9;
                while (!used.insert(g).second)
                {
                    g = rand() % 9;
                }

                w.m_layer[e] = hash<std::string>("default");
                w.m_scale[e] = gridToUnit * 0.8f;

                w.m_gridPos[e] = groupGridCentre + gridOffsets[g];

                w.m_pos[e] = groupCentre + gridOffsets[g] * gridToUnit;
                w.m_pos[e].y += gridToUnit.y * 0.4f;
            }
        }       

        // create ui canvas and children
        size_t uice = w.newEntity();
        w.m_layer[uice] = hash<std::string>("ui");
        loadEntity(uice, prefabs["combat"]["entities"]["ui_canvas"]);

        // create pause menu canvas and children
        size_t pme = w.newEntity();
        w.m_layer[pme] = hash<std::string>("ui");
        loadEntity(pme, prefabs["combat_pause_menu"]["entities"]["canvas"]);

        // now entities are loaded find layer cameras
        for (auto& l : m_layers)
        {
            if (l.second.m_cameraID == 0)
                continue;

            for (size_t k = 0, n = w.size(); k < n; ++k)
            {
                if (!(w.m_components[k] & e_cmp::camera))
                    continue;

                if (w.m_ulid[k] == (ULID)l.second.m_cameraID)
                {
                    l.second.m_cameraEntity = k;
                    break;
                }
            }
        }

        m_flags &= ~e_sceneFlags::loading;
        m_flags |= e_sceneFlags::loaded;
	}

	void CombatScene::activate()
	{
		Scene::activate();
	}

	void CombatScene::shutdown()
	{
		Scene::shutdown();
	}

	void CombatScene::update(float dt)
	{
        auto ctx = gfx::getMainContext();
        CombatWorld& w = *(CombatWorld*)m_world;

		Scene::update(dt);

        std::vector<View*> views;
        m_renderer->getViews(hash<std::string>("layer_post_process"), hash<std::string>("default"), views);

        for (auto& v : views)
        {
            if (v->m_pipeline != hash<std::string>("fog"))
                continue;

            // get the layer camera
            size_t cam = m_layers[hash<std::string>("default")].m_cameraEntity;

            glm::vec3 gridSize = w.m_gridScale[1];

            // update the data
            glm::vec4 falloffParams = glm::vec4(1.0f);
            falloffParams.x = gridSize.x;
            falloffParams.y = gridSize.y * 2.0f;
            falloffParams.z = gridSize.z;
            falloffParams.w = (float)utils::getSystemTimeMs() * 0.001f;
            *(glm::mat4*)v->m_uniformBufferMem.m_data = glm::inverse(w.m_camera[cam]->m_matrix);
            *(glm::vec4*)((intptr_t)v->m_uniformBufferMem.m_data + sizeof(glm::mat4)) = falloffParams;
            ctx->updateBuffer(v->m_uniformBuffer, v->m_uniformBufferMem);
            break;
        }
	}
}